﻿/*
 * Event types:
 * end_play
 * end_playcollect
 * end_playrecord
 * end_overlay
 * dtmf
 * tone
 * info
 * incoming
 * ringing
 * connected
 * hangup
 * keepalive 
 */


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace XmsDemo
{
    static class CallManager
    {

        private static Dictionary<string, XmsCall> m_callTable = new Dictionary<string, XmsCall>();
        private static bool m_isRunning = false;
        private static Mutex m_mutex = new Mutex();
        private static int rec_idx = 1;

        public static void Start()
        {
            if (!m_isRunning)
            {
                m_isRunning = true;
                Thread l_callCompletionThread = new Thread(onAsyncCallEvent);
               l_callCompletionThread.Start();
            }

        }

        public static void Stop()
        {
                m_isRunning = false;
        }

        public static int ProcessEvent(@event a_event)
        {
            XmsCall l_call = null;
//            Logger.Log("Before 1", true);
            m_mutex.WaitOne();
            
            try // looking for existing call
            {
                m_callTable.TryGetValue(a_event.resource_id, out l_call);
            }
            catch(ArgumentNullException aEx)
            {
                Logger.Log("ERROR Null key while searching the table : "+ aEx.Message, false);
            }
            m_mutex.ReleaseMutex();
//            Logger.Log("After 1", false);
            Logger.Log("Incoming event: "+ a_event.type.ToString() +
            " on resource " + a_event.resource_id, false);
 
            switch(a_event.type)
            {
                case event_type.incoming:
                    l_call = new XmsCall("");
                    l_call.CallId = a_event.resource_id;
                    OnOffered(ref l_call);
                    break;
                case event_type.connected:
                    OnConnect(ref l_call);
                    break;
                case event_type.dtmf:
                    OnDtmf(ref l_call);
                    break;
                case event_type.end_play:
                case event_type.end_playrecord:
                case event_type.end_playcollect:
                    OnEndPlay(ref l_call);
                    break;
                case event_type.hangup:
                    try
                    {
                        l_call = m_callTable[a_event.resource_id];
                    }
                    catch (KeyNotFoundException)
                    {
                        Logger.Log("ERR: Cannot find call reference to release", true);
                    }
                    if(l_call != null)
                        OnHangUp(l_call);
                    else
                        Logger.Log("ERR: Cannot find call reference to release", true);
                    break;
                case event_type.ringing:
                    break;
                case event_type.tone:
                    break;

            }
            return 0;
        }
        public static int MakeCall(string a_address)
        {
            XmsCall l_outCall = new XmsCall("");
            l_outCall.CallDirection = XmsCall.e_CallDirection.Outgoung;
            if (l_outCall.MakeCall(a_address) == -1)
                return -1;
            l_outCall.CallState = XmsCall.e_CallState.STATE_DIALING;
            addCall(ref l_outCall);
            return 0;

        }
// event handlers
        private static void OnOffered(ref XmsCall a_call)
        {
            a_call.CallState = XmsCall.e_CallState.STATE_OFFERED;
            a_call.CallDirection = XmsCall.e_CallDirection.Incoming;
            addCall(ref a_call);
            if (a_call.Answer() != 0) // failed for some reason, see logs, need to remove
            {
                a_call.Drop();
                deleteCall(a_call);
            }
        }
        private static void OnAnswer(XmsCall a_call)
        {
            
            Logger.Log("ANSWERED on call " + a_call.CallId, false);
            // just playing a message
            /*
            if (a_call.Play("file://verification/verification_intro.wav") == -1)
            {
                a_call.Drop();
            }
            */
            string l_recFile = new string('x', 1024);
            l_recFile = "playrec_" + rec_idx++; // +".wav";

            if (a_call.Record("file://verification/verification_intro.wav",
                "file://verification/" + l_recFile, "10s") == -1)
            {
                a_call.Drop();
            }

         }

        private static void OnConnect(ref XmsCall a_call)
        {
            Logger.Log("CONNECTED on call " + a_call.CallId, false);
            // just playing a message
            a_call.CallState = XmsCall.e_CallState.STATE_CONNECTED;

            if(a_call.Play("file://verification/verification_intro.wav") == -1)
            {
                a_call.Drop();
            }

        }

        private static void OnDrop(XmsCall a_call) // The app initiated call release
        {
            // Same code as below, just to separate local release from the remote one for clearness
            deleteCall(a_call);
        }

        private static void OnHangUp(XmsCall a_call) // usolicited release from the remote party
        {
            Logger.Log("Remote Hangup on call " + a_call.CallId, false);
//            a_call.Drop();
            
            deleteCall(a_call);
        }

        private static void OnDtmf(ref XmsCall a_call)
        {

        }
        private static void OnTone(ref XmsCall a_call)
        {

        }
        private static void OnKeepAlive(ref XmsCall a_call)
        {

        }
        private static void OnEndPlay(ref XmsCall a_call)
        {
            Logger.Log("Play or record complete on call " + a_call.CallId, false);
            a_call.Drop();
        }

        private static void onAsyncCallEvent()
        {
            LinkedList<XmsCall> l_callList = new LinkedList<XmsCall>();
            while(m_isRunning)
            {
                l_callList.Clear();
                m_mutex.WaitOne();
                foreach (KeyValuePair<string, XmsCall> kvp in m_callTable)
                    l_callList.AddLast(kvp.Value as XmsCall);
                m_mutex.ReleaseMutex();

                foreach(XmsCall l_call in l_callList)
                {
                    if (l_call.CallEvent.WaitOne(1))// completion event on this call
                    {
                        switch(l_call.CallState)
                        {
                            case XmsCall.e_CallState.STATE_FAIL:
                                deleteCall(l_call); // delete ther call
                                break;
                            case XmsCall.e_CallState.STATE_CONNECTED:
                                OnAnswer(l_call);
                                break;
                            case XmsCall.e_CallState.STATE_IDLE:
                                OnDrop(l_call);
                                break;
                        }
                            
                    }
                }
            }
        }

        private static void addCall(ref XmsCall a_call)
        {
//            Logger.Log("Before 2", true);
            m_mutex.WaitOne();
            try
            {
                m_callTable.Add(a_call.CallId, a_call);
            }
            catch (Exception ex)
            {
                Logger.Log("{ERROR: Cannot add a call to call table: " + ex.Message, true);
            }
            m_mutex.ReleaseMutex();
//            Logger.Log("After 2", false);
        }

        private static void deleteCall(XmsCall a_call)
        {
//            Logger.Log("Before 3", true);
            m_mutex.WaitOne();
            try
            {
                m_callTable.Remove(a_call.CallId);
            }
            catch (Exception ex)
            {
                Logger.Log("{ERROR: Cannot remove a call from call table: " + ex.Message, true);
            }
            m_mutex.ReleaseMutex();
//            Logger.Log("After 3", false);

        }
    }
}
