﻿
using System;
using System.Globalization;
using System.Xml;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Threading;
using System.Xml.Serialization;

namespace XmsDemo
{
    class EventHandler
    {
        private static string m_href; // referrence to the EH object, used for DELETing the handler
        private static string m_hId; // EH identifier
        private eventhandler m_EventHandler;
        private eventhandler_response m_EventResponse;
        private web_service m_ws;
        Thread m_eventThread;
        private static bool m_isRunning;
//        delegate void EventThread(int i);


        public int Create(string body)
        {
            m_EventHandler = new eventhandler();
            m_EventResponse = new eventhandler_response();
            m_ws = new web_service();
            return Subscribe(body);
        }

        public int Destroy()
        {
            return Unsubscribe();
        }


        private int Subscribe(string body) // sends POST method with event handler parameters
        {
            eventsubscribe[] es = new eventsubscribe[1];
            es[0] = new eventsubscribe();
            es[0].resource_id = "any";
            es[0].action = action_option.add;
            es[0].type = event_type.any;
            m_EventHandler.eventsubscribe = es;

            m_ws.Item = m_EventHandler as eventhandler;

            HttpWebRequest l_request = XmsInterface.CreateRequest("POST", "eventhandlers", "");
            try
            {
                using (Stream l_requestStream = l_request.GetRequestStream())
                {
                    XmlSerializer x = new XmlSerializer(typeof(web_service));
                    x.Serialize(l_requestStream, m_ws);
                }
            }
            catch (WebException ex)
            {
                Logger.Log("ERROR: " + ex.Message, false);
                return -1;
            }

            string l_srequest = XmsInterface.RequestToString(l_request, m_ws);
            Logger.Log(l_srequest, "Creating event handler", true);
            HttpWebResponse l_response = null;
            try
            {
                l_response = (HttpWebResponse)l_request.GetResponse();

                if (l_response.StatusCode != HttpStatusCode.Created)
                {
                    // process response here

                    Logger.Log("Failed to create event handler:\r\n" + XmsInterface.ResponseToString(l_response),
                         "Failed to create event handler", false);
                    return -1;
                }
                l_request.GetRequestStream().Close();
            }
            catch (Exception ex)
            {
                Logger.Log("ERROR: " + ex.Message, false);
                return -1;
            }

            using (Stream l_responseStream = l_response.GetResponseStream())
            {
                XmlSerializer x = new XmlSerializer(typeof(web_service));
                m_ws = (web_service)x.Deserialize(l_responseStream);
            }
            l_srequest = XmsInterface.ResponseToString(l_response);
            l_response.Close();
            

//            Logger.Log(l_srequest, false);
            Type tp = m_ws.Item.GetType();
            switch (tp.Name)
            {
                case "eventhandler_response":
                    eventhandler_response l_ehr = (eventhandler_response)m_ws.Item;
                    EventHandler.m_href = l_ehr.href;
                    EventHandler.m_hId = l_ehr.identifier;
                    Logger.Log(l_srequest, "Created event handler with ID " +
                        EventHandler.m_hId, false);
                    m_isRunning = true;
                    m_eventThread =  new Thread(EventHandler.EventThread);
                    m_eventThread.IsBackground = true;
                    m_eventThread.Start();
                    break;
            }
            return 0;
        }

        private int Unsubscribe()
        {
            m_isRunning = false;
            if (m_hId == null || m_hId == "")
                return -1;
            HttpWebRequest l_request = XmsInterface.CreateRequest("DELETE", "eventhandlers", m_hId);
            string l_srequest = XmsInterface.RequestToString(l_request, null);
            Logger.Log(l_srequest, "", true);

            HttpWebResponse l_response = (HttpWebResponse)l_request.GetResponse();
            l_srequest = XmsInterface.ResponseToString(l_response);
            Logger.Log(l_srequest, "",  false);
            return 0;
        }

        private static void EventThread()
        {
            Thread.Sleep(300);
            HttpWebRequest l_request = XmsInterface.CreateRequest("GET", "eventhandlers", m_hId);
            l_request.KeepAlive = true;
            l_request.Accept = null;
            l_request.ContentType = null;
            l_request.ProtocolVersion = HttpVersion.Version11;

            string l_srequest = XmsInterface.RequestToString(l_request, null);
            Logger.Log(l_srequest, "Starting event stream", true);

            HttpWebResponse l_response = (HttpWebResponse)l_request.GetResponse();
            l_srequest = "Status code: " + (int)l_response.StatusCode + " " + l_response.StatusCode;
            foreach (string key in l_response.Headers.Keys)
            {
                l_srequest += string.Format("{0}: {1} \r\n", key, l_response.Headers[key]);
            }
            if (l_response.StatusCode != HttpStatusCode.OK) // Failed
            {
                Logger.Log(l_srequest, "Failed to start event pump, see log", false);
                return;
            }

            StreamReader l_evtStream = new StreamReader(l_response.GetResponseStream());
            string line = "";
            while (m_isRunning)
            {
                {
                    try
                    {
                        line = l_evtStream.ReadLine(); // skipping unreadable
                        line = l_evtStream.ReadLine(); // reading event length
                        if (line == null || line.Length < 1) // may happen upon exit
                        {
                            continue;
                        }
                    }
                    catch (IOException ex)
                    {
                        Logger.Log("ERROR READ 2: " + ex.Message, false);
                        continue;
                    }
                    int buf_length = Int32.Parse(line, NumberStyles.AllowHexSpecifier); // parsing hex to int
                    char[] r_buf = new char[buf_length];
                    l_evtStream.Read(r_buf, 0, buf_length);

                    string str = new string(r_buf);
                    Logger.Log(str, "", false);
                    if(str.Length > 0) // not to pass empty payload
                        ProcessEvent(str);
                }
            }
        }

        private static void ProcessEvent(string a_xmlString)
        {
            
            XmlReader reader = XmlReader.Create(new StringReader(a_xmlString));

            web_service l_ws = new web_service();
            try
            {
                XmlSerializer x = new XmlSerializer(typeof(web_service));
                l_ws = (web_service)x.Deserialize(reader);
            }
            catch(Exception ex)
            {
                Logger.Log("ERROR @@: " + ex.Message, false);
                return;
            }

            Type tp = l_ws.Item.GetType();
            @event l_event = (@event)l_ws.Item;
            switch (l_event.resource_type)
            {
                case "call":
                    CallManager.ProcessEvent(l_event);
                    break;
                case "conference":
                    ConfManager.ProcessEvent(l_event);
                    break;
                case "keepalive":
                    Logger.Log("Keepalive from XMS", false);
                    break;
                default:
                    Logger.Log("ERROR - Unknown event type", false);
                    break;
            }

        }
 
    }

}
