﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Xml.Serialization;
using System.IO;
using System.Threading;


namespace XmsDemo
{
    class XmsCall
    {
        private call m_call;
        private call_action m_call_action;
        private play m_play;
        private playrecord m_playrec;
        private playcollect m_playcollect;
        HttpWebRequest m_request;
        private string m_callId;
        public string CallId
        {
            get{ return m_callId;}
            set{ m_callId = value;}
        }

        private string m_currentTransactionId; // for play/rec/overlay etc
        public string TransactionId
        {
            get { return m_currentTransactionId;}
            set { m_currentTransactionId = value; }
        }
        private AutoResetEvent m_asyncEvt;
        public AutoResetEvent CallEvent // event for signalling completion of ASYNC methods, like Answer and Drop
        {
            get { return m_asyncEvt; }
//            set { m_asyncEvt = value; }
        }

        public XmsCall(string a_callId)
        {
            this.CallId = a_callId;
            m_call = new call();
            m_call_action = new call_action();
            m_play = new play();
            m_playrec = new playrecord();
            m_playcollect = new playcollect();
            m_asyncEvt = new AutoResetEvent(false);
        }

        public enum e_CallState
        {
            STATE_OFFERED = 0, STATE_DIALING, STATE_CONNECTED, STATE_PLAYING,
            STATE_RECORDING, STATE_IDLE, STATE_NULL, STATE_FAIL
        };
        public enum e_CallDirection { Incoming, Outgoung }

        private e_CallState m_callState;
        private e_CallDirection m_callDirection;
        public e_CallState CallState
        {
            get { return m_callState; }
            set { m_callState = value; }
        }
        public e_CallDirection CallDirection
        {
            get { return m_callDirection; }
            set { m_callDirection = value; }
        }


        public int Answer()
        {
            m_call.answer = boolean_type.yes;
            m_call.answerSpecified = true;
            m_call.dtmf_mode = dtmf_mode_option.rfc2833;
            m_call.cpa = boolean_type.no;
            m_call.media = media_type.audio;
            m_call.async_dtmf = boolean_type.no;

            web_service l_ws = new web_service();
            l_ws.Item = m_call as call;
            
            try
            {
                m_request = XmsInterface.CreateRequest("PUT", "calls", this.CallId);
                using (Stream l_requestStream = m_request.GetRequestStream())
                {
                    XmlSerializer x = new XmlSerializer(typeof(web_service));
                    x.Serialize(l_requestStream, l_ws);
                }
                Logger.Log(XmsInterface.RequestToString(m_request, l_ws), "Answering call, cID:  " + m_callId, true);
// now waiting for a status from the server
                IAsyncResult asyncResult = m_request.BeginGetResponse(new AsyncCallback(onAnswer), this);
            }
            catch (WebException ex)
            {
                Logger.Log("ERROR: " + ex.Message, false);
                return -1;
            }
            return 0;
        }
        public int Drop()
        {
            try
            {
                m_request = XmsInterface.CreateRequest("DELETE", "calls", this.CallId);
                m_request.Accept = null;
                m_request.ContentType = null;

                Logger.Log(XmsInterface.RequestToString(m_request, null), "", true); // no XML pyload here
                IAsyncResult asyncResult = m_request.BeginGetResponse(new AsyncCallback(onDrop), this);
            }
            catch (WebException ex)
            {
                Logger.Log("ERROR: " + ex.Message, false);
                return -1;
            }
            return 0;
/*
    Here the SYNC way to get response from XMS:
            HttpWebResponse l_response = null;
            try
            {
                l_response = (HttpWebResponse)m_request.GetResponse();

            }
            catch (Exception ex)
            {
                Logger.Log("ERROR: " + ex.Message, false);
                return -1;
            }

            Logger.Log(XmsInterface.ResponseToString(l_response), "",  false);

            if (l_response.StatusCode != HttpStatusCode.NoContent)
            {
                // process response here
                Logger.Log("Failed to forward-release,Status " + l_response.StatusCode, false);
                return -1;
            }
            Logger.Log("The call released on resource " + this.CallId, false);
            l_response.Close();
            return 0;
 */ 
        }

        public int Play(string a_filename)
        {
            m_play.offset = "0s";
            m_play.repeat = "0";
            m_play.delay = "1s";
//            m_play.skip_interval = "0s";
//            m_play.max_time = "infinite";
            m_play.terminate_digits = "#";
            m_play.play_source = new play_source();
            m_play.play_source.location = a_filename;
            m_call_action.Item = m_play as play;
            m_call.call_action = m_call_action;
            web_service l_ws = new web_service();
            l_ws.Item = m_call as call;

            try
            {

                m_request = XmsInterface.CreateRequest("PUT", "calls", this.CallId);
                using (Stream l_requestStream = m_request.GetRequestStream())
                {
                    XmlSerializer x = new XmlSerializer(typeof(web_service));
                    x.Serialize(l_requestStream, l_ws);
                }
                Logger.Log(XmsInterface.RequestToString(m_request, l_ws), "Sending PLAY request on call " + this.CallId, true);
            }
            catch (WebException wex)
            {
                Logger.Log("ERROR: " + wex.Message, false);
            } 

            HttpWebResponse l_response = null;
            try
            {
                l_response = (HttpWebResponse)m_request.GetResponse();

            }
            catch (Exception ex)
            {
                Logger.Log("ERROR Playing file: " + ex.Message, false);
                return -1;
            }

            Logger.Log(XmsInterface.ResponseToString(l_response), "",  false);

            if (l_response.StatusCode != HttpStatusCode.OK)
            {
                Logger.Log("Failed to intialize playback on call " + this.CallId, false);
//                return -1;
            }
            else
                Logger.Log("Playing " + a_filename + " on call " + this.CallId, false);
            m_request.GetRequestStream().Close();
            l_response.Close();

            return 0;
        }
        public int Record(string a_playFileName, string a_recFileName, string a_durationInSec)
        {
/*
            if(a_playFileName == null)
            {
                m_playrec.barge = boolean_type.no;
            }
            else
                m_playrec.barge = boolean_type.yes;
            m_playrec.barge = boolean_type.yes;
*/
            m_playrec.barge = boolean_type.yes;
            m_playrec.beep = boolean_type.yes;
            m_playrec.delay = "0s";
            m_playrec.offset = "0s";
            m_playrec.cleardigits = boolean_type.yes;
            m_playrec.max_time = a_durationInSec;
            m_playrec.play_source = new play_source();
            m_playrec.play_source.location = a_playFileName;
            m_playrec.recording_uri = a_recFileName;
            m_playrec.repeat = "0";
            m_call_action.Item = m_playrec as playrecord;

            m_call.answerSpecified = false;
            m_call.cleardigitsSpecified = false;
            
            m_call.call_action = m_call_action;


            web_service l_ws = new web_service();
            l_ws.Item = m_call as call;

            try
            {
                m_request = XmsInterface.CreateRequest("PUT", "calls", this.CallId);
                using (Stream l_requestStream = m_request.GetRequestStream())
                {
                    XmlSerializer x = new XmlSerializer(typeof(web_service));
                    x.Serialize(l_requestStream, l_ws);
                }
                Logger.Log(XmsInterface.RequestToString(m_request, l_ws), "Sending PLAY-REC request on call " + this.CallId, true);
            }
            catch (WebException wex)
            {
                Logger.Log("ERROR: " + wex.Message, false);
            }

            HttpWebResponse l_response = null;
            try
            {
                l_response = (HttpWebResponse)m_request.GetResponse();

            }
            catch (Exception ex)
            {
                Logger.Log("ERROR Playing file: " + ex.Message, false);
                return -1;
            }

            Logger.Log(XmsInterface.ResponseToString(l_response), "", false);

            if (l_response.StatusCode != HttpStatusCode.OK)
            {
                Logger.Log("Failed to intialize playrec on call " + this.CallId, false);
                //                return -1;
            }
            else
                Logger.Log("Recording " + a_recFileName + " on call " + this.CallId, false);
            m_request.GetRequestStream().Close();
            l_response.Close();

            return 0;
        }

        public int MakeCall(string a_callingAddress)
        {
            web_service l_ws = new web_service();
            m_call.destination_uri = a_callingAddress;
            m_call.cpa = boolean_type.no;
            m_call.sdp = "";
            m_call.media = media_type.audio;
            m_call.signaling = boolean_type.yes;
            m_call.dtmf_mode = dtmf_mode_option.rfc2833;
            m_call.async_dtmf = boolean_type.no;
            m_call.async_dtmfSpecified = true; 
            m_call.async_tone = boolean_type.no;
            m_call.rx_delta = "+0dB";
            m_call.tx_delta = "+0dB";
            m_call.source_uri = "sip:" + XmsInterface.XmsUri();
            

            l_ws.Item = m_call as call;

            HttpWebRequest m_request = XmsInterface.CreateRequest("POST", "calls", "");
//            m_request.ProtocolVersion = HttpVersion.Version10;

            using (Stream l_requestStream = m_request.GetRequestStream())
            {
                XmlSerializer x = new XmlSerializer(typeof(web_service));
                x.Serialize(l_requestStream, l_ws);
            }

            Logger.Log(XmsInterface.RequestToString(m_request, l_ws),
                "Making call to " + a_callingAddress, true);

            HttpWebResponse l_response = null;
            try
            {
                l_response = (HttpWebResponse)m_request.GetResponse();
            }
            catch (Exception ex)
            {
                Logger.Log("ERROR making call: " + ex.Message, false);
                return -1;
            }

            if (l_response.StatusCode != HttpStatusCode.Created)
            {
                // process response here
                Logger.Log(XmsInterface.ResponseToString(l_response), "MakeCall failed", false);
                return -1;
            }

            using (Stream l_responseStream = l_response.GetResponseStream())
            {
                XmlSerializer x = new XmlSerializer(typeof(web_service));
                l_ws = (web_service)x.Deserialize(l_responseStream);
            }

            Logger.Log(XmsInterface.ResponseToString(l_response), "",  false);

            m_request.GetRequestStream().Close();

            Type tp = l_ws.Item.GetType();
            switch (tp.Name)
            {
                case "call_response":
                    call_response l_cr = (call_response)l_ws.Item;
                    this.m_callId = l_cr.identifier;
                    Logger.Log("Waiting for answer from  " + a_callingAddress + 
                " on call ID " + this.m_callId, false);
                    break;
                default:
                    Logger.Log("Unknown response object type of " + tp.Name, false);
                    break;
            }
            try
            {
                m_request.GetRequestStream().Close();
                l_response.Close();
            }
            catch (NullReferenceException ex)
            {
                Logger.Log("ERROR closing streams: " + ex.Message, true); 
            }
            return 0;
        }

// callbacks 
        private static void onAnswer(IAsyncResult a_result)
        {
            XmsCall l_call = (XmsCall)a_result.AsyncState;
            HttpWebRequest l_request = null;
            HttpWebResponse l_response = null;
            try
            {
                l_request = l_call.m_request;
                l_response = (HttpWebResponse)l_request.EndGetResponse(a_result);
                Logger.Log(XmsInterface.ResponseToString(l_response), "", false);
                if (l_response.StatusCode != HttpStatusCode.OK)
                {
                    Logger.Log("Failed to answer call with ID " + l_call.m_callId, false); // add more processing here
                    l_call.Drop();
//                    l_call.CallState = e_CallState.STATE_FAIL;
                }
                else
                {
                    l_call.CallState = e_CallState.STATE_CONNECTED;
//                    Logger.Log("Answered call with ID " + l_call.m_callId, false); // add more processing here
                }
            }
            catch (WebException e)
            {
                Logger.Log("ERROR: " + e.Message, false);
                l_call.Drop();
            }
            l_call.CallEvent.Set(); // signal to listeners about completion of ASYNC operation
            if(l_request != null)
                l_request.GetRequestStream().Close();
            if(l_response != null)
                l_response.Close();
        }


        private static void onDrop(IAsyncResult a_result)
        {
            XmsCall l_call = (XmsCall)a_result.AsyncState;
            HttpWebRequest l_request = null;
            HttpWebResponse l_response = null;

            try
            {
                l_request = l_call.m_request;
                l_response = (HttpWebResponse)l_request.EndGetResponse(a_result);
                Logger.Log(XmsInterface.ResponseToString(l_response), "", false);
                if (l_response.StatusCode != HttpStatusCode.NoContent)
                {
                    Logger.Log("Failed to drop call with ID " + l_call.m_callId, false); // add more processing here
                    l_call.CallState = e_CallState.STATE_FAIL;
                }
                else
                {
                    l_call.CallState = e_CallState.STATE_IDLE;
                    Logger.Log("Released call with ID " + l_call.m_callId, false); // add more processing here
                }

            }
            catch (WebException e)
            {
                Logger.Log("ERROR: " + e.Message, false);
                l_call.CallState = e_CallState.STATE_FAIL;
            }
            l_call.CallEvent.Set(); // signal to listeners about completion of ASYNC operation
            if (l_request != null)
                l_request.GetRequestStream().Close();
            if (l_response != null)
                l_response.Close();
        }
    }
}
